#include "timelinelistmodel.h"
#include <QDebug>
#include <QSqlQuery>
#include <QVariant>
#include "database.h"
#include "timelinemodel.h"

using namespace Qt::Literals::StringLiterals;

struct TimelineListModel::Item
{
    Item(int id, const QString name, QObject *parent = nullptr)
        : id(id)
        , name(name)
        , timeline(new TimelineModel(parent))
    {}

    ~Item() { timeline->deleteLater(); }

    int id;
    QString name;
    TimelineModel *timeline;
};

TimelineListModel::TimelineListModel(QObject *parent)
    : QAbstractListModel{parent}
    , m_items{}
    , m_fromDate{}
    , m_toDate{}
{}

TimelineListModel::~TimelineListModel()
{
    qDeleteAll(m_items);
}

QVariant TimelineListModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (role != Qt::DisplayRole) {
        return {};
    }
    return section;
}

int TimelineListModel::rowCount(const QModelIndex &parent) const
{
    if (parent.isValid()) {
        return 0;
    }

    return m_items.count();
}

QHash<int, QByteArray> TimelineListModel::roleNames() const
{
    QHash<int, QByteArray> roles = QAbstractItemModel::roleNames();
    roles[NameRole] = "name";
    roles[TimelineRole] = "timeline";
    return roles;
}

void TimelineListModel::clear()
{
    if (m_items.isEmpty()) {
        return;
    }
    beginResetModel();
    qDeleteAll(m_items);
    m_items.clear();
    endResetModel();
}

QDate TimelineListModel::fromDate() const
{
    return m_fromDate;
}

void TimelineListModel::setFromDate(QDate date)
{
    if (date == m_fromDate) {
        return;
    }
    m_fromDate = date;
    emit fromDateChanged(m_fromDate);
    refill();
}

QDate TimelineListModel::toDate() const
{
    return m_toDate;
}

void TimelineListModel::setToDate(QDate date)
{
    if (date == m_toDate) {
        return;
    }
    m_toDate = date;
    emit toDateChanged(m_toDate);
    refill();
}

QModelIndex TimelineListModel::findIndexById(int id)
{
    for (qsizetype row = 0; row < m_items.count(); ++row) {
        if (m_items.at(row)->id == id) {
            return index(row);
        }
    }
    return {};
}

void TimelineListModel::refill()
{
    if (!m_fromDate.isValid() || !m_toDate.isValid()) {
        clear();
        return;
    }

    struct Lodging
    {
        int id;
        QString label;
    };

    struct Results
    {
        QList<Lodging> lodgings;
        QHash<int, QList<TimelineModel::Item *>> bookings;
    };

    QDate fromDate = m_fromDate;
    QDate toDate = m_toDate;
    Database::query([fromDate, toDate]() {
        Results results;
        {
            QSqlQuery query("select campsite_id, label from campsite order by label, campsite_id");
            Database::checkError(query);
            while (query.next()) {
                Lodging item{query.value(0).toInt(), query.value(1).toString()};
                results.lodgings.append(item);
            }
        }
        {
            QSqlQuery query;
            query.setForwardOnly(true);
            query.prepare(" select "
                          "    campsite_id "
                          "  , booking_id "
                          "  , lower(booking_campsite.stay) "
                          "  , upper(booking_campsite.stay) - lower(booking_campsite.stay)"
                          "  , holder_name "
                          "  , booking_status "
                          "  , true"
                          "  , true"
                          " from booking_campsite "
                          " join booking using (booking_id) "
                          " where booking_campsite.stay && daterange(:fromDate, :toDate)"
                          " order by lower(booking_campsite.stay), booking_id "
                          ""_L1);
            Database::checkError(query);
            query.bindValue(":fromDate"_L1, fromDate);
            query.bindValue(":toDate"_L1, toDate);
            query.exec();
            Database::checkError(query);
            while (query.next()) {
                int lodgingId = query.value(0).toInt();
                auto *item = new TimelineModel::Item{
                    query.value(1).toInt(),
                    query.value(2).toDate(),
                    query.value(3).toInt(),
                    query.value(4).toString(),
                    query.value(5).toString(),
                    query.value(6).toBool(),
                    query.value(7).toBool(),
                };
                results.bookings[lodgingId].append(item);
            }
        }
        return results;
    }).then(this, [this](const Results &results) {
        qsizetype itemsRow = 0;
        qsizetype lodgingsRow = 0;

        for (; itemsRow < m_items.count() && lodgingsRow < results.lodgings.count();
             ++itemsRow, ++lodgingsRow) {
            Item *item = m_items.at(itemsRow);
            const Lodging &lodging = results.lodgings.at(lodgingsRow);

            if (lodging.id == item->id) {
                if (lodging.label != item->name) {
                    item->name = lodging.label;
                    QModelIndex itemIndex = index(itemsRow);
                    emit dataChanged(itemIndex, itemIndex, {Qt::DisplayRole, NameRole});
                };
                continue;
            }

            if (QModelIndex itemIndex = findIndexById(lodging.id); itemIndex.isValid()) {
                beginMoveRows({}, itemIndex.row(), itemIndex.row(), {}, itemsRow);
                m_items.move(itemIndex.row(), itemsRow);
                endMoveRows();
                // Move one step backwards, and check change of label in next loop.
                --itemsRow;
                --lodgingsRow;
            } else {
                beginInsertRows({}, itemsRow, itemsRow);
                m_items.insert(itemsRow, new Item(lodging.id, lodging.label, this));
                endInsertRows();
            }
        }
        if (itemsRow < m_items.count()) {
            beginRemoveRows({}, itemsRow, m_items.count() - 1);
            qDeleteAll(m_items.begin() + itemsRow, m_items.end());
            m_items.remove(itemsRow, m_items.count() - itemsRow);
            endRemoveRows();
        }
        if (lodgingsRow < results.lodgings.count()) {
            beginInsertRows({},
                            m_items.count(),
                            m_items.count() + results.lodgings.count() - lodgingsRow - 1);
            for (; lodgingsRow < results.lodgings.count(); ++lodgingsRow) {
                const Lodging &lodging = results.lodgings.at(lodgingsRow);
                m_items.append(new Item(lodging.id, lodging.label, this));
            }
            endInsertRows();
        }
        for (Item *item : m_items) {
            item->timeline->update(results.bookings.value(item->id));
        }
    });
}

QVariant TimelineListModel::data(const QModelIndex &index, int role) const
{
    if (!checkIndex(index,
                    QAbstractItemModel::CheckIndexOption::IndexIsValid
                        | QAbstractItemModel::CheckIndexOption::ParentIsInvalid)) {
        return {};
    }

    Item *item = m_items.at(index.row());
    switch (role) {
    case Qt::DisplayRole:
    case NameRole:
        return item->name;
    case TimelineRole:
        return QVariant::fromValue(item->timeline);
    }

    return {};
}

#include "moc_timelinelistmodel.cpp"
